/*********************************************

        CES-11 - LAB 03 - Bitmap quadtrees

        Este pedao de programa implementa a
        leitura e a escrita de imagens preto
        e branco em arquivos BMP do Windows
        
         necessrio converter a imagem no
        Paint para 2 cores (preto e branco)
        para poder utilizar nesse programa.
        
        autor: Carlos Henrique Q. Forster
        Instituto Tecnolgico de Aeronutica
        2010
*********************************************/

#include <stdio.h>
#include <stdlib.h>

/* emite mensagem de erro e aborta */
void erro(char *msg)
{
    fprintf(stderr,"\nERRO: %s\n",msg);
    abort();
}

/* ler um inteiro (de 2 bytes little-endian) do arquivo binrio*/
int readint(FILE *fp)
{
    int x;
    x=getc(fp)+256*getc(fp);
    return x;
}

/* ler um long (de 4 bytes) do arquivo binrio */
long readlong(FILE*fp)
{
    long x;
    x=getc(fp);
    x+=256*getc(fp);
    x+=256*256*getc(fp);
    x+=256*256*256*getc(fp);
    return x;
}

/* pular certo nmero de bytes do arquivo binrio */
void skipbytes(FILE *fp, int count)
{
    int i;
    for (i=0; i<count; i++) getc(fp);
}

/* leitura de um bit (buffer de um byte)*/
unsigned bitpack; int avbits=0;
int readbit(FILE *fp)
{
    int x;
    if (avbits==0)
    {
        bitpack=getc(fp);
        avbits=8;
    }
    x=(bitpack & 0x0080)!=0;
    bitpack=(bitpack<<1) & 0x00FE;
    avbits--;
    return x;
}
        
/* escrever um inteiro (de 2 bytes little-endian) no arquivo binrio*/
void writeint(int val, FILE *fp)
{
    putc(val%256,fp);
    putc(val/256,fp);
}

/* escrever um long (de 4 bytes) no arquivo binrio */
void writelong(long val, FILE*fp)
{
    putc(val%256,fp); val/=256;
    putc(val%256,fp); val/=256;
    putc(val%256,fp); val/=256;
    putc(val%256,fp);
}

/* preencher com zeros certo nmero de bytes do arquivo binrio */
void fillbytes(FILE *fp, int count)
{
    int i;
    for (i=0; i<count; i++) putc(0,fp);
}

/* escrita de um bit */
unsigned wrbitpack=0; int wravbits=0;
void writebit(int x,FILE *fp)
{
    if (wravbits==8)
    {
        putc(wrbitpack,fp);
        wrbitpack=0;
        wravbits=0;
    }
    wrbitpack=(wrbitpack <<1)|(x!=0);
    wravbits++;
}

/* terminar escritas/leituras pendentes dos bits */
void flushbit(FILE *fp)
{
    if(wravbits)
    {
        wrbitpack= wrbitpack <<(8-wravbits);
        putc(wrbitpack,fp);
        wravbits=0;
    }
    bitpack=avbits=0;
}
        
/* ler um arquivo de bitmap binario para memoria */
/* retorna uma estrutura matricial de apontadores */
/* deve se passar o nome do arquivo, uma variavel */
/* por referencai para obter a largura e outra */
/* para receber a altura */
char **readbitmap(char *filename, int *width, int *height)
{
    int type,planes,bits;
    long size,offset,fixedwidth,compression,imsize,xres,yres;

    int i, j;
    FILE *fp;
    
    char **imagem;

    fp=fopen(filename,"rb");
    type=readint(fp); if(type!=19778) erro("nao eh BMP");
    size=readlong(fp);
    skipbytes(fp,4); // reserved
    offset=readlong(fp);
    skipbytes(fp,4); // structure size
    *width=readlong(fp);
    fixedwidth=32*(1+(*width-1)/32); //armazenada em multiplos de 32 bits
    *height=readlong(fp);
    planes=readint(fp); // always 1
    bits=readint(fp); if(bits!=1) erro("nao eh binaria");// 1 4 8 or 24
    compression=readlong(fp);
    imsize=readlong(fp); //size of image in bytes
    xres=readlong(fp);
    yres=readlong(fp);
    skipbytes(fp,8); // all zero
    
    skipbytes(fp,8); // the two colors

    flushbit(fp);
    imagem=(char **) malloc(sizeof(char*)* (*height) );
    for(i=0; i<*height; i++)
    {
        imagem[*height-i-1]=(char *) malloc(sizeof(char)* (*width));
        for(j=0; j<*width; j++) imagem[*height-i-1][j]=readbit(fp);
        for(j=*width; j<fixedwidth; j++) readbit(fp);
    } // a figura fica de cabea para baixo na imagem

    fclose(fp);
    return imagem;
}


/* escrever um bitmap em arquivo BMP */
void writebitmap(char * filename, char **imagem, int width, int height)
{
    long fixedwidth, offset, imsize, size;
    int i, j;
    FILE *fp;
    
    fixedwidth=32*(1+(width-1)/32);
    offset=62;
    imsize=(fixedwidth*height)/8;
    size=offset+imsize;
    
    fp=fopen(filename,"wb");
    writeint(19778,fp);
    writelong(size,fp);
    fillbytes(fp,4); // reserved
    writelong(offset,fp); // offset
    writelong(40,fp); // structure size
    writelong(width,fp);
    writelong(height,fp);
    writeint(1,fp); // planes=always 1
    writeint(1,fp); // bits=1
    writelong(0,fp); // compression=0
    writelong(imsize,fp); //size of image in bytes
    writelong(0,fp); // xres=0
    writelong(0,fp); // yres=0
    fillbytes(fp,8); // all zero
    
    fillbytes(fp,4);
    putc(0x0FF,fp); putc(0x0FF,fp); putc(0x0FF,fp); putc(0x00,fp);// the two colors
    for(i=0; i<height; i++)
    {
            for(j=0; j<width; j++) writebit(imagem[height-i-1][j],fp);
            for(j=width; j<fixedwidth; j++) writebit(0,fp);
    } // a figura fica de cabea para baixo
    flushbit(fp);

    fclose(fp);
}


/* criar uma imagem em memria para editar e salvar */
char **createbitmap(int width, int height)
{
    char **imagem;
    int i,j;
    
    imagem=(char **) malloc(sizeof(char*)*height);
    for(i=0; i<height; i++)
    {
        imagem[i]=(char *) malloc(sizeof(char)*width);
        for(j=0; j<width; j++) imagem[i][j]=1; // preenche de branco
    }
    return imagem;
}    


/* liberar uma imagem da memria*/
void freebitmap(char **img, int width, int height)
{
    int i;
    
    for(i=0; i<height; i++)
    {
        free(img[i]);
    }
    free(img);
}    



main()
{
    char **img, **img2;
    
    int t;
    int i,j,ww,hh;
    
    img=createbitmap(60,50);
    for(t=0;t<50; t++)
      img[t][((t-30)*(t-30)*49)/900]=0;
    writebitmap("parabola.bmp",img,60,50);
    freebitmap(img,60,50);
        
    img=readbitmap("um.bmp",&ww,&hh);
    writebitmap("dois.bmp",img,ww,hh);
    freebitmap(img,ww,hh);
    
    img=readbitmap("semibola.bmp",&ww,&hh);
    img2=createbitmap(ww,hh*2);
    for (i=0; i<hh; i++)
      for(j=0; j<ww; j++)
        img2[i][j]=img2[2*hh-i-1][j]=img[hh-i-1][j];
    writebitmap("bola.bmp",img2,ww,2*hh);
    freebitmap(img,ww,hh);
    freebitmap(img2,ww,2*hh);
    
    img=readbitmap("pagina.bmp",&ww,&hh);
    for (i=0; i<hh; i++)
      for (j=0; j<ww; j++)
        img[i][j]=1-img[i][j];
    writebitmap("negpagina.bmp",img,ww,hh);
    freebitmap(img,ww,hh);
    
    printf("Pronto!!");

    getchar(); getchar();
}
    
